// SPDX-FileCopyrightText: 2026 Sandro Andrade <sandroandrade@kde.org>
//
// SPDX-License-Identifier: GPL-2.0-or-later

#pragma once

namespace Minuet
{
void showIosSplashScreen();
void hideIosSplashScreen();
}
